:not(pre)>code[class*=language-], pre[class*=language-], div[class*="language-"], .theme-default-content pre, .theme-default-content pre[class*=language-]
  background #F5F5F6

div[class*="language-"] .highlight-lines .highlighted
  background-color rgba(0, 0, 0, 0.08)

.theme-default-content pre code, .theme-default-content pre[class*="language-"] code
  color #000000  !important

.token.atrule, .token.builtin, .token.important, .token.keyword, .token.selector
  color #A9261B  !important

div[class*="language-"]::before
  color #999 !important

.token.punctuation
  color #C58D09 !important

.token.entity, .token.operator, .token.url
  color #004898  !important

.token.boolean, .token.number, .token.function
  color #4C10BC !important

.token.string, .token.char, .token.attr-value, .token.regex, .token.variable
  color #429E9E !important

.token.property, .token.class-name, .token.constant, .token.symbol
  color #67BF89 !important
