@font-face
  font-family: 'Basis Grotesque Pro'
  font-style: normal
  font-weight: normal
  src: local('Basis Grotesque Pro'), url('/biome-text/v3.0.0/assets/fonts/BasisGrotesquePro-Regular.woff') format('woff')


@font-face
  font-family: 'Basis Grotesque Pro Bold'
  font-style: normal
  font-weight: normal
  src: local('Basis Grotesque Pro Bold'), url('/biome-text/v3.0.0/assets/fonts/BasisGrotesquePro-Bold.woff') format('woff')


@font-face
  font-family: 'Basis Grotesque Pro Light'
  font-style: normal
  font-weight: normal
  src: local('Basis Grotesque Pro Light'), url('/biome-text/v3.0.0/assets/fonts/BasisGrotesquePro-Light.woff') format('woff')
