// font family
@import url('https://fonts.googleapis.com/css2?family=Open+Sans&display=swap')
@require "./fonts.styl"
$primaryFontFamily = 'Basis Grotesque Pro'
$secondaryFontFamily = 'Open Sans'
$handMadeFontFamily = 'Just Me Again Down Here', cursive

// colors
$accentColor = #F38959
$textColorLight = #686A6D
$textColor = #4A4A4A
$borderColor = #D8D8D8
$codeBgColor = #4A4A4A
$arrowBgColor = #ccc
$badgeTipColor = #9013FE
$badgeWarningColor = darken(#ffe564, 35%)
$badgeErrorColor = #DA5961
$sidebarBgColor = #F5F5F6
$codePillColor = #F0E7FF
$yellow = #F8D11C
$green = #6ACE91
$red = #FF1E5E

// layout
$navbarHeight = 3.6rem
$sidebarWidth = 20rem
$contentWidth = 740px
$homePageWidth = 960px

// responsive breakpoints
$MQNarrow = 959px
$MQMobile = 719px
$MQMobileNarrow = 419px
